from __future__ import annotations
from typing import Any, Union
import enum
import copy
import functools
from typing import Tuple
from ..utils import almost_equal
from .Complex import Complex
from .Vector import Vector
from .Field import Field, DefaultRealField
from ..utils import areinstances, check_foreach, isoneof

# TODO remove hardcoded solution vec and add it as parameter to solve with


class MatrixOperationType(enum.Enum):
    ROW_SWITCHING = "row_switching"
    ROW_MULTIPLICATION = "row_multiplication"
    ROW_ADDITION = "row_addition"
    COL_SWITCHING = "column_switching"
    COL_MULTIPLICATION = "column_multiplication"
    COL_ADDITION = "column_addition"


MOT = MatrixOperationType


# class MatrixOperation:
#     def __init__(self, operation_type: MatrixOperationType, iv1: Union[int, float], iv2: Union[int, float]):
#         self.operation_type = operation_type
#         self.iv1, self.iv2 = iv1, iv2

#     def __call__(self, matrix, operate_with: Union[Vector, Matrix] = None) -> Tuple[Matrix, Union[Vector, Matrix]]:
#         res = matrix
#         iv1, iv2 = self.iv1, self.iv2
#         match(self.operation_type):
#             case MatrixOperationType.ROW_SWITCHING:
#                 res[iv1], res[iv2] = res[iv2], res[iv1]
#             case MatrixOperationType.COL_SWITCHING:
#                 for i in range(len(res[0])):
#                     res[i][iv1], res[i][iv2] = res[i][iv2], res[i][iv1]
#             case MatrixOperationType.ROW_MULTIPLICATION:
#                 res[iv1] = [iv2*v for v in res[iv1]]
#             case MatrixOperationType.COL_MULTIPLICATION:
#                 for i in range(len(res[0])):
#                     res[i][iv1] *= iv2
#             case MatrixOperationType.ROW_ADDITION:
#                 res[iv1] = [res[iv1][i]+res[iv2][i]
#                             for i in range(len(res[iv1]))]
#             case MatrixOperationType.COL_ADDITION:
#                 for i in range(len(res[0])):
#                     res[i][iv1] += res[i][iv2]
#         return res, operate_with


class Matrix:

    @staticmethod
    def fromVector(vec: Vector, sol: Vector = None) -> Matrix:
        """
        will add the vector as a column
        """
        # return Matrix.fromVectors([vec])
        return Matrix([[v] for v in vec], sol)

    @staticmethod
    def fromVectors(vecs: list[Vector], sol: Vector = None) -> Matrix:
        """
        will create a amtrix from the vectors in the order they appear and as columns
        """
        if not areinstances(vecs, Vector):
            raise TypeError("all elements must be instances of class 'Vector'")
        if not check_foreach(vecs, lambda v: v.field == vecs[0].field):
            raise ValueError("vectors are not over the same field")
        mat: list[list[Any]] = []
        for i in range(vecs[0].length):
            mat.append([])
            for j in range(len(vecs)):
                mat[i].append(vecs[j][i])
        return Matrix(mat, sol, field=vecs[0].field)

    @staticmethod
    def fromString(matrix_string: str, sol: Vector) -> Matrix:
        return Matrix([[int(num) for num in row.split()]
                       for row in matrix_string.split("\n")], sol)

    @staticmethod
    def random(rows: int, cols: int, f: Field = None, min: float = -10, max: float = 10, degree: int = 10,  def_value=None) -> Matrix:
        if f is None:
            f = DefaultRealField
        # TODO how to check that defualt value is inside 'f'? what if 'f' is ratinals and has no __contains__ implemented?
        return Matrix([[f.random(min, max) if def_value is None else def_value for _ in range(cols)]for __ in range(rows)], field=f)

    @staticmethod
    def fromJordanBlocks(lst: list[Matrix]) -> Matrix:
        pass

    @staticmethod
    def createJordanBlock(size: int, eigenvalue) -> Matrix:
        pass

    @staticmethod
    def id_matrix(size: int) -> Matrix:
        arr = [[0 for __ in range(size)] for _ in range(size)]
        for i in range(size):
            arr[i][i] = 1
        return Matrix(arr)

    def __init__(self, mat: list[list[Any]], sol_vec: list[Union[float, Complex]] = None, field: Field = DefaultRealField) -> None:
        if not isinstance(mat, list) or not all([isinstance(v, list) for v in mat]):
            raise TypeError("Matrix must be a 2d array")
        self.__matrix = mat
        self.__rows = len(mat)
        self.__cols = len(mat[0])
        self.__solution_vector = sol_vec if sol_vec else [
            0 for _ in range(self.__rows)]
        self.field = field

    @property
    def kernel(self) -> list[Vector]:
        solution = Matrix(
            self.__matrix, [0 for _ in range(self.__rows)]).solve()

    @property
    def image(self) -> list[Vector]:
        # TODO implement image calculation
        pass

    @property
    def rank(self) -> int:
        # TODO fix rank calculation this is not correct
        self.reorgenize_rows()
        rank = 0
        for row in self.__matrix:
            if all([i == 0 for i in row]):
                continue
            rank += 1
        return rank

    @property
    def determinant(self) -> float:
        if self.__rows != self.__cols:
            raise ValueError("Matrix must be square")
        if self.__rows == 1:
            return self.__matrix[0][0]
        if self.__rows == 2:
            return self.__matrix[0][0] * self.__matrix[1][1] - self.__matrix[0][1] * self.__matrix[1][0]
        return sum([self.__matrix[i][0] * ((-1)**i) * self.minor(i, 0) for i in range(self.__rows)])

    @property
    def is_invertiable(self) -> bool:
        return self.determinant != 0

    @property
    def is_square(self) -> bool:
        return self.__rows == self.__cols

    @property
    def is_diagonialable(self) -> bool:
        # TODO implement diagonialability check
        pass

    @property
    def is_nilpotent(self) -> bool:
        # TODO implement nilpotency check
        pass

    @property
    def eigen_values(self) -> Vector:
        # TODO implement eigen value calculation
        pass

    @property
    def jordan_form(self) -> Matrix:
        # TODO implement jordan form calculation
        pass

    @property
    def chain_basis(self) -> list[Vector]:
        # TODO implement chain basis calculation
        pass

    # @property
    # def characteristic_polynomial(self) -> SimplePolynomial.SimplePolynomial:
    #     # TODO implement characteristic polynomial calculation
    #     pass

    # @property
    # def minimal_polynomial(self) -> SimplePolynomial.SimplePolynomial:
    #     # TODO implement minimal polynomial calculation
    #     pass

    def __getitem__(self, index: int) -> Any:
        if not isinstance(index, int):
            raise TypeError("Index must be an integer")
        return self.__matrix[index]

    def __setitem__(self, index: int, value: list[Any]) -> None:
        if not isinstance(index, int):
            raise TypeError("Index must be an integer")
        if not isinstance(value, list):
            raise TypeError("Value must be a list")
        if not 0 <= index < self.__rows:
            raise ValueError("Index out of range")
        # TODO validate all items in list are of same type of this matrix
        self.__matrix[index] = value

    def __str__(self, turnc: int = 2) -> str:
        def find_spaceing():
            res = 0
            for row in self:
                for v in row:
                    if hasattr(v, '__round__'):
                        if almost_equal(round(v), v):
                            v = round(v)
                    res = max(res, len(str(v)))
            return res
        spacing = find_spaceing()+2
        n = len(self[0])
        vl = "|"
        hl = (1+spacing)*n*"-" + "\n"
        result = hl
        for i, row in enumerate(self.__matrix):
            result += vl
            for v in row:
                if hasattr(v, '__round__'):
                    if almost_equal(round(v), v):
                        v = round(v)
                result += str(v).center(spacing)+vl
            result += "\n"+hl
            # result += " | "+str(self.__solution_vector[i]) + "\n"
        return result

    def __add__(self, other: Matrix) -> Matrix:
        if not isinstance(other, Matrix):
            raise TypeError("Matrix can only be added to another Matrix")
        if self.__rows != other.__rows or self.__cols != other.__cols:
            raise ValueError("Matrices must have the same dimensions")
        return Matrix([[self.__matrix[i][j] + other.__matrix[i][j] for j in range(self.__cols)]
                       for i in range(self.__rows)])

    def __neg__(self) -> Matrix:
        return Matrix([[-self.__matrix[i][j] for j in range(self.__cols)]
                       for i in range(self.__rows)])

    def __sub__(self, other: Matrix) -> Matrix:
        if not isinstance(other, Matrix):
            raise TypeError(
                "Matrix can only be subtracted from another Matrix")
        if self.__rows != other.__rows or self.__cols != other.__cols:
            raise ValueError("Matrices must have the same dimensions")
        return Matrix([[self.__matrix[i][j] - other.__matrix[i][j] for j in range(self.__cols)]
                       for i in range(self.__rows)])

    def __mul__(self, other: Union[float, Complex, Vector, Matrix]) -> Union[float, Complex, Vector, Matrix]:
        """
        self * other
        """
        from ..la2 import Calculable
        if isoneof(other, [int, float, Complex, Calculable]):
            res: list[list[Any]] = []
            # res = self
            for i in range(len(self)):
                res.append([])
                for j in range(len(self[0])):
                    res[i].append(self[i][j] * other)
            return Matrix(res)
            # return Matrix([[other * self.__matrix[i][j] for j in range(len[self[0]])]
            #                for i in range(len(self))])
        if isinstance(other, Vector):
            if self.__cols != other.length:
                raise ValueError(
                    "Matrix and Vector must have the same number of rows")
            return Vector([sum([self.__matrix[i][j] * other[j] for j in range(self.__cols)])
                           for i in range(self.__rows)])
        if isinstance(other, Matrix):
            if self.__cols != other.__rows:
                raise ValueError(
                    "Matrix and Matrix must have matching sizes: self.cols == other.rows")
            return Matrix([[sum([self.__matrix[i][j] * other.__matrix[j][k] for j in range(self.__cols)])
                            for k in range(other.__cols)] for i in range(self.__rows)])

        raise TypeError(
            "Matrix can only be multiplied by a number, Vector, or Matrix")

    def __rmul__(self, other: Union[int, float, Complex, Vector, Matrix]) -> Union[float, Complex, Vector, Matrix]:
        """
        other * self
        """
        if isoneof(other, [int, float, Complex]):
            return self.__mul__(other)
        if isinstance(other, Vector):
            raise TypeError(
                "Matrix can only be multiplied by a vector from the right")
        if isinstance(other, Matrix):
            if self.__cols != other.__rows:
                raise ValueError(
                    "Matrix and Matrix must have the same number of columns")
            return Matrix([[sum([self.__matrix[i][j] * other.__matrix[j][k] for j in range(self.__cols)])
                            for k in range(other.__cols)] for i in range(self.__rows)])
        raise TypeError(
            f"cant perform {type(other)}*Matrix.\ncan only be multiplied by a:\n\tint\n\tfloat\n\tComplex\n\tVector\n\tMatrix]")

    def __eq__(self, other: Matrix) -> bool:
        if not isinstance(other, Matrix):
            raise TypeError(f"cant complare 'Matrix' with '{type(other)}'")
        if self.__rows != other.__rows or self.__cols != other.__cols:
            return False
        if any([self.__matrix[i][j] != other.__matrix[i][j] for i in range(self.__rows) for j in range(self.__cols)]):
            return False
        return True

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __len__(self) -> int:
        return self.__rows

    def __pow__(self, other) -> Matrix:
        if isoneof(other, [int, float]):
            if other == int(other):
                other = int(other)
                res = self
                for _ in range(other-1):
                    res *= self
                return res
            raise NotImplementedError(
                "Matrix**float is not implemented")
        raise NotImplementedError(
            f"Matrix power not implemented for Matrix**{type(other)}")

    def almost_equal(self, other: Matrix) -> bool:
        if not isinstance(other, Matrix):
            raise TypeError("Matrix can only be compared to another Matrix")
        if not self.field == other.field:
            raise ValueError("Matrix must have the same field")
        return all([all([almost_equal(self[i][j], other[i][j])] for j in range(len(self[0]))) for i in range(len(self))])

    def inverse(self) -> Matrix:
        if not self.is_invertiable:
            raise ValueError("Matrix must be invertible")
        return Matrix([[self.minor(i, j) / self.determinant for j in range(self.__cols)]
                       for i in range(self.__rows)])

    def cofactor(self, row_to_ignore: int, col_to_ignore: int) -> Matrix:
        if(row_to_ignore >= self.__rows or col_to_ignore >= self.__cols):
            raise ValueError("Row or column index out of range")
        res: list[list[Any]] = []
        for i, row in enumerate(self.__matrix):
            if i == row_to_ignore:
                continue
            res.append([])
            for j, col in enumerate(self.__matrix[i]):
                if j == col_to_ignore:
                    continue
                res[i if i < row_to_ignore else i -
                    1].append(self.__matrix[i][j])
        return Matrix(res)

    def minor(self, row_to_ignore: int, col_to_ignore: int) -> float:
        return self.cofactor(row_to_ignore, col_to_ignore).determinant

    def transpose(self) -> Matrix:
        return Matrix([[self.__matrix[j][i] for j in range(self.__cols)]
                       for i in range(self.__rows)])

    def reorgenize_rows(self):
        def comparer(a: list[float], b: list[float]) -> bool:
            def first_not_zero_index(row: list[float]) -> int:
                for i in range(len(row)):
                    if row[i] != 0:
                        break
                return i
            return -1 if first_not_zero_index(a) > first_not_zero_index(b) else 1
        self.__matrix = sorted(
            self.__matrix, key=functools.cmp_to_key(comparer), reverse=True)

    def apply_operation(self, operation: MatrixOperationType, iv1, iv2, operate_with=None) -> Matrix:
        if not isinstance(operation, MatrixOperationType):
            raise TypeError("can only apply MatrixOperationType")
        match(operation):
            case MOT.ROW_MULTIPLICATION:
                self[iv1] = [iv2*v for v in self[iv1]]
            case MOT.ROW_ADDITION:
                self[iv1] = [self[iv1][i]+self[iv2][i]
                             for i in range(len(self[iv1]))]
            case MOT.ROW_SWITCHING:
                self[iv1], self[iv2] = self[iv2], self[iv1]
            case MOT.COL_SWITCHING:
                for i in range(len(self[0])):
                    self[i][iv1], self[i][iv2] = self[i][iv2], self[i][iv1]
            case MOT.COL_MULTIPLICATION:
                for i in range(len(self[0])):
                    self[i][iv1] *= iv2
            case MOT.COL_ADDITION:
                for i in range(len(self[0])):
                    self[i][iv1] += self[i][iv2]
        return operate_with

    def concat(self, other) -> Matrix:
        if not isoneof(other, [Vector, Matrix]):
            raise TypeError("can only concat with a Vector or Matrix")
        if len(other) != len(self):
            raise TypeError(
                "can only concat with a Matrix|Vector with the same number of rows")
        if isinstance(other, Vector):
            other = Matrix.fromVector(other)
        res = []
        for i in range(len(self)):
            res.append(self[i] + other[i])
        return Matrix(res)

    def split(self, index: int) -> Tuple[Matrix, Matrix]:
        if not isinstance(index, int):
            raise TypeError("can only split at an int")
        if not 0 <= index <= len(self[0]):
            raise ValueError("index out of range")
        if index == len(self[0]):
            return self, None
        return Matrix([[self[i][j] for j in range(index)] for i in range(len(self))], self.field), Matrix([[self[i][j] for j in range(index, len(self[0]))] for i in range(len(self))], self.field)

    def guassian_elimination(self, solve_with: Union[Vector, Matrix] = None) -> Tuple[Matrix, Union[None, Matrix]]:
        SPLIT_INDEX = len(self[0])
        res = self
        if solve_with is not None:
            if not (isoneof(solve_with, [Vector, Matrix]) or solve_with is None):
                raise TypeError("can only solve with a Vector or Matrix")
            res = self.concat(solve_with)
        # reorder rows
        res.reorgenize_rows()
        # performe gaussian elimination
        for curr_row_index in range(len(res)):
            # find lead value for current row
            for lead_index in range(curr_row_index, SPLIT_INDEX):
                if res[curr_row_index][lead_index] != 0:
                    break
            else:
                break
            lead_value = res[curr_row_index][lead_index]
            # make lead value equal one and change row acordingly
            if lead_value != 1:
                res.apply_operation(MOT.ROW_MULTIPLICATION,
                                    curr_row_index, 1/lead_value)
            for next_row_index in range(len(res)):
                # skip current row
                if next_row_index == curr_row_index:
                    continue
                # find if need to operate
                for next_row_lead_index in range(lead_index, SPLIT_INDEX):
                    if res[next_row_index][next_row_lead_index] != 0:
                        break
                else:
                    continue
                if next_row_lead_index == lead_index:
                    row_multiplyer = res[next_row_index][next_row_lead_index]
                    res.apply_operation(
                        MOT.ROW_MULTIPLICATION, next_row_index, -1/row_multiplyer)
                    res.apply_operation(
                        MOT.ROW_ADDITION, next_row_index, curr_row_index)
                    res.apply_operation(
                        MOT.ROW_MULTIPLICATION, next_row_index, -row_multiplyer)
        return res.split(SPLIT_INDEX)

    def vectorize(self) -> Vector:
        arr = []
        for col in range(len(self[0])):
            arr += [self[row][col] for row in range(len(self))]
        return Vector(arr)

    def solve(self, vec: Vector) -> Union[Vector, list[Vector], None]:
        """
        Solve the system of equations
        """
        if self.__rows != self.__cols:
            raise ValueError("Matrix must be square")
        if not isinstance(vec, Vector):
            raise TypeError("Matrix must be solved for a vector")
        result_matrix, sol = self.guassian_elimination(vec)
        sol = sol.vectorize()
        # if rank of matrix is equal to number of rows and cols return solution vector
        if result_matrix.rank == len(result_matrix):
            return sol
        # if there is a row in result matrix which has all zeros and solution vector doesnt has a zero at that row there is no solution
        for i, row in enumerate(result_matrix):
            if (not Vector(row).has_no_zero) and sol[i] != 0:
                return None
        # otherwise there is a span of solutions
        # TODO implement solutions for nullity rank > 1
        raise NotImplementedError(
            "nullity rank was atleast 1, not yet implemented")

    def get_eigen_vectors_of(eigenvalue) -> list[Vector]:
        # TODO calculate the eigen space of an eigen value
        pass

    def algebraic_multiplicity(eigenvalue) -> int:
        # TODO calculate the algebraic multiplicity of an eigenvalue
        pass

    def geometric_multiplicity(eigenvalue) -> int:
        # TODO calculate the geometric multiplicity of an eigenvalue
        pass

    # def guassian_elimination(self, sol=None) -> Matrix:
    #     if not sol:
    #         sol = self.__solution_vector

    #     def first_not_zero_index(row: list[float]) -> int:
    #         for i in range(len(row)):
    #             if row[i] != 0:
    #                 break
    #         return i
    #     res = copy.deepcopy(self)
    #     res.__solution_vector = sol
    #     res.reorgenize_rows()
    #     # gaussian elimination
    #     for r in range(res.__rows):
    #         lead_index = first_not_zero_index(res[r])
    #         lead_value = res[r][lead_index]
    #         if lead_value == 0:
    #             continue
    #         if lead_value != 1:
    #             for c in range(res.__cols):
    #                 res[r][c] *= 1/lead_value
    #             res.__solution_vector[r] *= 1/lead_value
    #             lead_value = res[r][lead_index]
    #         for r2 in range(res.__rows):
    #             if r == r2:
    #                 continue
    #             row_divider = res[r2][lead_index]/lead_value
    #             if row_divider == 0:
    #                 continue
    #             for c in range(res.__cols):
    #                 res[r2][c] -= row_divider * res[r][c]
    #             res.__solution_vector[r2] -= row_divider * \
    #                 res.__solution_vector[r]
    #     return res



from ...la1 import Field, Complex as c, RealField, ComplexField, DefaultComplexField, DefaultRealField, DefaultRationalField


def test_is_field():
    assert Field.is_field(DefaultComplexField)
    assert Field.is_field(DefaultRealField)
    assert Field.is_field(DefaultRationalField)


def test_clone():
    r3 = RealField(3)
    r21 = r3.classOfInstance(2)
    r23 = r3.classOfInstance.create(r3._name, 2)
    assert r21 == r23



from ..PolynomialSimple import PolynomialSimple as p
from ....la1 import Matrix

x = p([1], [1])
xp1 = p([1, 1], [1, 0])
square = p([1], [2])
sqrt = p([1], [1/2])


def test_constuctor_special_case():
    assert p([1, 2], [1, 1]) == p([3], [1])


def test_addition():
    assert x+5 == p([1, 5], [1, 0])
    assert 5+x == p([1, 5], [1, 0])
    assert 5+x+5 == p([1, 10], [1, 0])
    assert x+p([2], [2]) == p([1, 2], [1, 2])


def test_with_Field():
    from ....la1 import Vector, LinearMap, RealField, VectorSpace
    field = RealField(2)

    def func(v, target_field):
        return Vector([0, v[0]], target_field)
    lt = LinearMap(field, field, func)
    v = VectorSpace(field).random()
    # assert Polynomial.fromString("x^2")(lt)(v) == Vector([0, 0], field)
    assert p.fromString("x^2+1")(lt)(v) == v
    assert p.fromString(
        "x+1")(lt)(v) == Vector([v[0], sum(v)], field)


def test_subtraction():
    assert x-5 == p([1, -5], [1, 0])
    assert 5-x == p([5, -1], [0, 1])
    assert 5-x-5 == p([-1], [1])
    assert x-p([2], [2]) == p([1, -2], [1, 2])
    assert x-x == 0


def test_negation():
    assert -x == p([-1], [1])


def test_multiplication():
    assert 1*x == x
    assert x*5 == p([5], [1])
    assert x*x == p([1], [2])
    assert xp1*xp1 == p([1, 2, 1], [2, 1, 0])
    assert p.fromString(
        "(x^2+2*x+5)")*p.fromString("1") == p.fromString("1")*p.fromString("(x^2+2*x+5)")


def test_with_matrix():
    from ....la1 import Matrix
    assert p.fromString("x^2")(
        Matrix([[1, 0], [0, 1]])) == Matrix([[1, 0], [0, 1]])


def test_call():
    assert x(5) == 5
    assert x(5.0) == 5.0
    assert xp1(5) == 6
    assert x*xp1(2) == p([3], [1])
    assert (x*x)(5) == 25
    assert (xp1*xp1)(2) == (2+1)**2
    assert square(square)(2) == 16
    assert square(xp1)(2) == (2+1)**2
    assert sqrt(4) == 2


# def test_division():
#     assert x/1 == (x, 0)
#     # TODO finish implementing so tests will pass
#     # assert 1/x == p([1], [-1])
#     assert xp1/x == (p([1], [0]), p([1], [-1]))


# test_division()


def test_pow():
    assert x**0 == 1
    assert x**1 == x
    assert x**2 == x*x
    assert x**3 == x*x*x
    assert (xp1**2)(2) == (2+1)**2
    assert (xp1**2)(2.0) == (2+1)**2
    assert x**-1 == 1/x


def test_equals():
    assert p([0], [0]) == 0 == p([0], [534273])
    assert p([], []) == 0
    assert p([0], []) == 0
    assert p([1], [0]) == 1


def test_fromstring():
    f = p.fromString
    assert f("x^2 +2x+1") == p([1, 2, 1], [2, 1, 0])
    assert f(
        "-x^7+x^6-34*x^5-x4-x^3+x^2") == p([-1, -1, 34, -1, -1, -4], [7, 6, 5, 3, 2, 1])
    assert f("(x+1)") == p([1, 1], [1, 0])
    assert f("((x+1))(x+1)") == p([1, 2, 1], [2, 1, 0]) == f("(x+1)^2")
    # TODO can run faster need to improve this
    assert f("((1)(x^2+2*x+5))^2") == f("X^4 + 4X^3 + 14X^2 + 20X + 25", "X")


def test_with_matrix():
    m = Matrix.id_matrix(2)
    px = p.fromString("x")



from __future__ import annotations
from typing import Callable, Any, Union
from .Field import Field
from ..utils import composite_function, isoneof
from .Matrix import Matrix
from .Vector import Vector
from .Complex import Complex
from .VectorSpace import VectorSpace


class LinearMap:
    # TODO rethink data structure
    @staticmethod
    def isFuncLinearTransformation(func: Callable[[Any, Field], Union[Vector, Matrix]], src_field: Field, dst_field: Field) -> bool:
        COUNT = 100
        V = VectorSpace(src_field)
        for _ in range(COUNT):
            a, b = src_field.random(), src_field.random()
            v1, v2 = V.random(), V.random()
            try:
                if not func(a*v1+b*v2, dst_field).almost_equal(a*func(v1, dst_field)+b*func(v2, dst_field)):
                    pass
            except Exception as e:
                pass
            if not func(a*v1+b*v2, dst_field).almost_equal(a*func(v1, dst_field)+b*func(v2, dst_field)):
                return False
        return True

    @staticmethod
    def fromMatrix(m: Matrix) -> LinearMap:
        return LinearMap(m.field, type(m.field)(m.field._name), lambda x: m*x)

    @staticmethod
    def id(field: int) -> LinearMap:
        return LinearMap(field, field, lambda x, y: x)

    def __init__(self, src_field: Field, dst_field: Field, func: Callable[[Any], Any], validate: bool = False) -> None:
        """creates a new linear transformation

        Args:
            src_field (Field): The source field which elements form it are called with this opeartor
            dst_field (Field): The Field which is the output oif this transformation
            func (Callable[[Any], Any]): the transformation function
        """
        if validate:
            if not LinearMap.isFuncLinearTransformation(func, src_field, dst_field):
                raise ValueError("func is not a linear transformation")
        self.src_field = src_field
        self.dst_field = dst_field
        self.func = func

    def __add__(self, other) -> LinearMap:
        # if isoneof(other, [int, float, complex]):
        # return LinearTransformation(self.src_field, self.dst_field, lambda x, y: self.func(x, y)+other*LinearTransformation())
        if isinstance(other, LinearMap):
            if self.src_field == other.src_field and self.dst_field == other.dst_field:
                return LinearMap(self.src_field, self.dst_field, lambda x, y: self(x)+other(x))
            raise ValueError(
                "cant add linear transformations on diffrent fields")
        raise NotImplementedError(
            "addition with non-numeric type not implemented")

    def __radd__(self, other) -> LinearMap:
        return self.__add__(other)

    def __sub__(self, other) -> LinearMap:
        return self.__add__(-other)

    def __rsub__(self, other) -> LinearMap:
        return other+(-self)

    def __neg__(self) -> LinearMap:
        return self.__mul__(-1)

    def __mul__(self, other) -> LinearMap:
        if isoneof(other, [int, float, Complex]):
            return LinearMap(self.src_field, self.dst_field, lambda x, y: self.func(x, y)*other)
        else:
            raise NotImplementedError(
                "multiplication with non-numeric type not implemented")

    def __rmul__(self, other) -> LinearMap:
        return self.__mul__(other)

    def __pow__(self, other) -> LinearMap:
        if isoneof(other, [int, float]):
            if not other == int(other) or other < 0:
                raise ValueError(
                    "only non negativ integer powers are implemented and you tried to raise the transformation to {}".format(other))
            if other == 0:
                return LinearMap(self.src_field, self.dst_field, lambda x, y: x)
            if other == 1:
                return self

            other = int(other)
            func = composite_function(self.func, self.func)
            for _ in range(abs(other)-2):
                func = composite_function(func, self.func)
            return LinearMap(self.src_field, self.dst_field, lambda x, y: func(x, y))
        else:
            raise NotImplementedError(
                "multiplication with non-numeric type not implemented")

    def __truediv__(self, other) -> LinearMap:
        # TODO
        pass

    def __call__(self, v: Union[Vector, Matrix]) -> Union[Vector, Matrix]:
        """ apply the transformation on an object
        Args:
            v (Any): the object to apply the transformation on

        Raises:
            ValueError: will raise a ValueError if the object is not in the source field
            Exception: if the transformation function itself raises an error

        Returns:
            Any: the return value of the transformation's function
        """
        if not isoneof(v, [Vector, Matrix]):
            raise ValueError(
                "can only apply linear transformations on vectors and matrices")
        if isinstance(v, Vector):
            if not v.field == self.src_field:
                raise ValueError("value is not in the source field")
        else:
            if not Vector(v[i][0] for i in range(v.__rows)).field == self.src_field:
                raise ValueError("value is not in the source field")
        try:
            return self.func(v, self.dst_field)
        except Exception as e:
            raise e

    def toMatrix(self, base=None) -> Matrix:
        # TODO implement calculation of operator over specific base
        pass


class LinearTransformation(LinearMap):
    pass



from __future__ import annotations
from typing import Any, Tuple
from ...utils import isoneof, areinstances
from ...la1 import Complex
from .PolynomialSimple import PolynomialSimple


class PolynomialFraction:
    @staticmethod
    def fromString(string, var="x") -> PolynomialFraction:
        pass

    @staticmethod
    def fromPolynomials(nominator: PolynomialSimple, denominator: PolynomialSimple) -> PolynomialFraction:
        if not areinstances([nominator, denominator], PolynomialSimple):
            raise TypeError("nominator and denominator must be Polynomials")
        return PolynomialFraction(nominator.prefixes, nominator.powers, denominator.prefixes, denominator.powers)

    @staticmethod
    def fromPrimitive(value) -> PolynomialFraction:
        if not isoneof(value, [int, float, Complex]):
            raise TypeError("value must be a number")
        return PolynomialFraction([value], [0])

    def __init__(self, nominator_prefixes: list, nominator_powers: list, denominator_prefixes: list = [1], denominator_powers: list = [0]) -> None:
        self.nominator = PolynomialSimple(nominator_prefixes, nominator_powers)
        self.denominator = PolynomialSimple(
            denominator_prefixes, denominator_powers)

    def __add__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            n1, d1 = self.destructure()
            n2, d2 = other.destructure()
            return PolynomialFraction.fromPolynomials(n1*d2+n2*d1, d1*d2)
        elif isinstance(other, PolynomialSimple):
            return self+PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self+PolynomialFraction.fromPrimitive(other)
        raise NotImplementedError(
            f"Addition of Expressions with {type(other)} not implemented yet")

    def __radd__(self, other) -> PolynomialFraction:
        return self.__add__(other)

    def __sub__(self, other) -> PolynomialFraction:
        return self + (-other)

    def __rsub__(self, other) -> PolynomialFraction:
        return other + (-self)

    def __neg__(self) -> PolynomialFraction:
        return self.__mul__(-1)

    def __mul__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            return PolynomialFraction.fromPolynomials(self.nominator*other.nominator, self.denominator*other.denominator)
        elif isinstance(other, PolynomialSimple):
            return self*PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self*PolynomialFraction.fromPrimitive(other)
        raise NotImplementedError(
            "Multiplication of Expressions with {} not implemented yet".format(type(other)))

    def __rmul__(self, other) -> PolynomialFraction:
        return self.__mul__(other)

    def __truediv__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            n1, d1 = self.destructure()
            n2, d2 = other.destructure()
            return PolynomialFraction.fromPolynomials(n1*d2, d1*n2)
        elif isinstance(other, PolynomialSimple):
            return self / PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self / PolynomialFraction.fromPrimitive(other)
        raise NotImplementedError(
            "Division of Expressions with {} not implemented yet".format(type(other)))

    def __rtruediv__(self, other) -> PolynomialFraction:
        # if isinstance(other, Expression):
        #     n1, d1 = self.destructure()
        #     n2, d2 = other.destructure()
        #     return Expression.fromPolynomials(n1*d2, d1*n2)
        if isinstance(other, PolynomialSimple):
            return self / PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return PolynomialFraction.fromPrimitive(other) / self
        raise NotImplementedError(
            "Division of Expressions with {} not implemented yet".format(type(other)))

    def __eq__(self, other) -> PolynomialFraction:
        if isinstance(other, PolynomialFraction):
            n1, d1 = self.destructure()
            n2, d2 = other.destructure()
            # n1/d1 == n2/d2 <==> n1d2/d1d2 == n2d1/d1d2 <==> n1d2 == n2d1
            return n1*d2 == n2*d1
        elif isinstance(other, PolynomialSimple):
            return self == PolynomialFraction(other.prefixes, other.powers)
        elif isoneof(other, [int, float, Complex]):
            return self == PolynomialFraction.fromPrimitive(other)
        return False

    def __ne__(self, other) -> PolynomialFraction:
        return not self.__eq__(other)

    def __pow__(self, other) -> PolynomialFraction:
        if isinstance(other, int):
            return PolynomialFraction.fromPolynomials(self.nominator**other, self.denominator**other)
        raise NotImplementedError(
            "Exponentiation of Expressions not implemented yet")

    def __rpow__(self, other) -> PolynomialFraction:
        raise NotImplementedError(
            "Exponentiation of Expressions not implemented yet")

    def __str__(self) -> str:
        if self.denominator != 1:
            return f"({str(self.nominator)})/({str(self.denominator)})"
        return str(self.nominator)

    def __call__(self, value) -> Any:
        if isoneof(value, [int, float, Complex, PolynomialSimple, PolynomialFraction]):
            try:
                return self.nominator(value) / self.denominator(value)
            except ArithmeticError as e:
                raise e
        else:
            raise ValueError(
                "PolynomialFraction.__call__() called with invalid argument")

    def destructure(self) -> Tuple[PolynomialSimple, PolynomialSimple]:
        return self.nominator, self.denominator

    def simplify() -> PolynomialFraction:
        pass
# from __future__ import annotations
# from functools import reduce
# import Polynomial
# from typing import Union, Callable, Any, Tuple
# import Complex
# from utils import isoneof, alloneof, validate_brackets
# import Vector
# import LinearTransformation
# class Expression:
#     @staticmethod
#     def fromString(s: str, var: str = "x") -> Expression:
#         if not validate_brackets(s):
#             raise ValueError("invalid brackets")
#         if var not in s:
#             raise ValueError("Variable not found in string")
#         if s.count(var) > 1:
#             raise NotImplementedError(
#                 "Not implemented more than one appearance of variable")
#         if s.count("^") > 1:
#             raise NotImplementedError(
#                 "Not implemented more than one appearance of power")
#         if s.count("*") > 1:
#             raise NotImplementedError(
#                 "Not implemented more than one appearance of multiplication")
#         if s.count("/") > 1:
#             raise NotImplementedError(
#                 "Not implemented more than one appearance of division")
#         if s.count("+") > 1:
#             raise NotImplementedError(
#                 "Not implemented more than one appearance of addition")
#         if s.count("-") > 1:
#             raise NotImplementedError(
#                 "Not implemented more than one appearance of subtraction")
#         def extract(string: str, variable: str, operator: str) -> Tuple[str, float]:
#             tmp = string.split(operator)
#             try:
#                 if len(tmp) == 2:
#                     remainder, num = None, None
#                     if var in tmp[0]:
#                         num = float(tmp[1])
#                         remainder = tmp[0]
#                     else:
#                         num = float(tmp[0])
#                         remainder = tmp[1]
#                     return remainder.strip(), num
#                 else:
#                     if len(tmp) == 1 and variable in tmp[0]:
#                         return tmp[0].strip(), 1 if operator in ["*", "^"] else 0
#                     elif len(tmp) == 1 and variable not in tmp[0]:
#                         return "", float(tmp[0])
#                     assert False, "shouldnt not be here"
#             except ValueError as e:
#                 raise e

#         try:
#             if "(" not in s:
#                 remainder, b = extract(s, var, "+")
#                 remainder, a = extract(remainder, var, "*")
#                 _, p = extract(remainder, var, "^")
#                 return Expression(a, p, b, lambda x: a*(x**p)+b)
#             else:
#                 if s.count("(") > 1:
#                     raise NotImplementedError(
#                         "Not implemented more than one appearance of brackets")
#                 inner_as_text = s[s.find("(")+1:s.find(")")]
#                 inner = Expression.fromString(inner_as_text, var)
#                 wrapper = Expression.fromString(
#                     s.replace(f"({inner_as_text})", "x"), var)
#                 return Expression.fromFunction(lambda x: wrapper(inner(x)))
#         except Exception as e:
#             raise e

#     @ staticmethod
#     def fromFunction(f: Callable[[Any], Any]) -> Expression:
#         return Expression(None, None, None, f)

#     def __init__(self, a=1, p=1, b=0, func=None) -> None:
#         """_summary_
#         Initialize an Expression object
#         Args:
#             a (_type_): _description_
#             p (_type_): _description_
#             b (_type_): _description_
#         """
#         if alloneof([a, p, b, func], [type(None)]):
#             raise ValueError("Cannot create Expression with all None")
#         self.a = a
#         self.p = p
#         self.b = b
#         self.func = func

#     @ property
#     def derivetive(self) -> Expression:
#         if not alloneof([self.a, self.p, self.b], [type(None)]):
#             return Expression(self.a*self.p, self.p-1, 0, lambda x: self.a*self.p*(x**(self.p-1)))
#         else:
#             raise ValueError("cant be calculated")

#     @ property
#     def antiderivative(self) -> Expression:
#         if not alloneof([self.a, self.p, self.b], [type(None)]):
#             return Expression(self.a/(self.p+1), self.p+1, 0, lambda x: self.a/(self.p+1)*(x**(self.p+1)))
#         else:
#             raise ValueError("cant be calculated")

#     @property
#     def minimal_polyinomial(self) -> Polynomial.Polynomial:
#         pass

#     def __add__(self, other: Union[int, float, Complex.Complex, Expression, Polynomial.Polynomial]) -> Union[Expression, Polynomial.Polynomial]:
#         if isinstance(other, Expression):
#             if self.func and other.func:
#                 return Expression.fromFunction(lambda x: self(x) + other(x))
#         elif isoneof(other, [int, float, Complex.Complex]):
#             if self.func:
#                 return Expression.fromFunction(lambda x: self(x) + other)
#         raise NotImplementedError("Not implemented")

#     def __radd__(self, other) -> Union[Expression, Polynomial.Polynomial]:
#         return self.__add__(other)

#     def __sub__(self, other) -> Expression:
#         return self.__add__(-other)

#     def __rsub__(self, other) -> Union[Expression, Polynomial.Polynomial]:
#         return other + (-self)

#     def __neg__(self) -> Expression:
#         if alloneof([self.a, self.p, self.b], [type(None)]):
#             return Expression.fromFunction(lambda x: -self(x))
#         return Expression(-self.a, self.p, self.b, lambda x: -self.func(x) if self.func else None)

#     def __truediv__(self, other) -> Expression:
#         if isinstance(other, Expression):
#             if self.func and other.func:
#                 return Expression.fromFunction(lambda x: self.func(x)/other.func(x))

#     def __mul__(self, other: Union[int, float, Complex.Complex, Expression, Polynomial.Polynomial]) -> Union[Expression, Polynomial.Polynomial]:
#         if isinstance(other, Expression):
#             if self.func and other.func:
#                 return Expression.fromFunction(lambda x: self(x) * other(x))
#         elif isoneof(other, [int, float, Complex.Complex]):
#             if self.func:
#                 return Expression.fromFunction(lambda x: self(x) * other)
#         raise NotImplementedError(
#             f"Not implemented Expression * {type(other).__name__}")

#     def __rmul__(self, other) -> Expression:
#         return self.__mul__(other)

#     def __pow__(self, power) -> Expression:
#         if isoneof(power, [int, float]):
#             if self.func:
#                 return Expression.fromFunction(lambda x: self(x)**power)
#         raise NotImplementedError(
#             f"Not implemented Expression ^ {type(power).__name__}")

#     def __rpow__(self, base) -> Expression:
#         if isinstance(base, Expression):
#             if self.func and base.func:
#                 return Expression.fromFunction(lambda x: base(x)**self(x))
#         elif isoneof(base, [int, float, Complex.Complex]):
#             if self.func:
#                 return Expression.fromFunction(lambda x: base**self(x))
#         raise NotImplementedError(
#             f"Not implemented {type(base).__name__} ^ Expression")

#     def __call__(self, x: Union[int, float, Complex.Complex]) -> Union[int, float, Complex.Complex]:
#         if alloneof([self.a, self.p, self.b], [type(None)]):
#             return self.func(x)
#         if alloneof([self.a, self.b, self.p], [int, float, Complex.Complex]):
#             if isoneof(x, [int, float, Complex.Complex]):
#                 if isinstance(self.p, Complex.Complex):
#                     raise NotImplementedError("Complex power not implemented")
#                 return self.func(x) if self.func else self.a * (x**self.p)+self.b
#             else:
#                 NotImplementedError("Not implemented")
#         else:

#             raise NotImplementedError("Not implemented")

#     def __str__(self) -> str:
#         if not alloneof([self.a, self.p, self.b], [type(None)]):
#             if self.a == 0:
#                 return f"{self.b}"
#             if self.b == 0:
#                 if self.p == 0:
#                     return f"{self.a}"
#                 else:
#                     if self.a == 1:
#                         return f"x^{self.p}"
#                 return f"{self.a}*x^{self.p}"
#             return f"{self.a}*x^{self.p}+{self.b}"
#         return "No representations for lambda"

#     def toVector(self) -> Vector.Vector:
#         # TODO rethink if this needs to be here or in Polynomial.Polynomial and if so, how to do it
#         pass



from __future__ import annotations
from .PolynomialFraction import PolynomialFraction


class Expression:
    @staticmethod
    def fromString(string: str) -> Expression:
        pass

    def __init__(self, string): pass



from __future__ import annotations
from enum import Enum
import random
from ..utils import are_operators_implemnted, almost_equal
from .Complex import Complex
from typing import Any, Callable


class Fields(Enum):
    Q = "Q"
    R = "R"
    C = "C"
    M = "M"


class Field:
    @staticmethod
    def create(name: Fields, degree: int = 1, modulu: int = 1) -> Field:
        match name:
            case Fields.Q:
                return RationalField(degree, modulu)
            case Fields.R:
                return RealField(degree, modulu)
            case Fields.C:
                return ComplexField(degree, modulu)
            case Fields.M:
                # TODO Field create function for Fields.M
                pass
                # return MatrixField(name, zero, one, degree, modulu)

    @staticmethod
    def is_field(field: Field) -> bool:
        N = 100
        if not are_operators_implemnted(type(field.random())):
            raise NotImplementedError(
                "One of the nescesary operators for calculation was not implemented")

        def checker(var_count: int, rule: Callable[[], bool], exclude=None) -> bool:
            if exclude is None:
                exclude = []
            for _ in range(N):
                vars = [field.random() for _ in range(var_count)]
                for v in vars:
                    if v in exclude:
                        break
                else:
                    if not rule(*vars):
                        pass
                    if not rule(*vars):
                        return False
            return True

        def associativity() -> bool:
            def addition() -> bool:
                return checker(3, lambda a, b, c: almost_equal((a+b)+c, a+(b+c)))

            def multiplication() -> bool:
                return checker(3, lambda a, b, c: almost_equal((a*b)*c, a*(b*c)))
            return all([addition(), multiplication()])

        def commutativity() -> bool:
            def addition() -> bool:
                return checker(2, lambda a, b: almost_equal(a+b, b+a))

            def multiplication() -> bool:
                return checker(2, lambda a, b: almost_equal(a*b, b*a))
            return all([addition(), multiplication()])

        def distributivity() -> bool:
            def addition() -> bool:
                return checker(3, lambda a, b, c: almost_equal(a*(b+c), a * b + a * c))

            def multiplication() -> bool:
                return checker(3, lambda a, b, c: almost_equal((a+b)*c, a * c + b * c))
            return all([addition(), multiplication()])

        def identity() -> bool:
            def addition() -> bool:
                return checker(1, lambda a: almost_equal(a+field._zero, a, field._zero+a))

            def multiplication() -> bool:
                return checker(1, lambda a: almost_equal(a*field._one, a, field._one*a))
            return all([addition(), multiplication()])

        def inverses() -> bool:
            def addition() -> bool:
                return checker(1, lambda a: almost_equal(a + (-a), field._zero, (-a)+a))

            def multiplication() -> bool:
                return checker(1, lambda a: almost_equal(a * (1/a), field._one, (1/a)*a), [0])
            return all([addition(), multiplication()])
        return all([associativity(), commutativity(), distributivity(), identity(), inverses()])

    def __init__(self, name: Fields,  zero, one, degree: int = 1, modulu: int = 1, validate: bool = False) -> None:
        if not isinstance(name, Fields):
            raise TypeError("'name' must be from enum 'Fields'")
        if not isinstance(degree, int):
            raise TypeError("'degree' must be of type 'int'")
        if not isinstance(modulu, int):
            raise TypeError("'modulu' must be of type 'int'")
        self._name = name
        self._modulu = modulu
        self._degree = degree
        self._zero = zero
        self._one = one
        if validate:
            if not Field.is_field(self):
                raise ValueError(
                    "This is not a field as one or more of the axioms do not check-out")

    @property
    def classOfInstance(self) -> Field:
        # raise NotImplementedError("This is a virtual method")
        return type(self)

    def __str__(self) -> str:
        return f"{self._name.value}{self._degree}%{self._modulu}"

    def __eq__(self, other: Field) -> bool:
        return self._name == other._name and self._modulu == other._modulu and self._degree == other._degree and self._zero == other._zero and self._one == other._one

    # def __contains__(self, obj: Any) -> bool:
    #     """
    #     """
    #     raise NotImplementedError("This is a virtual method")

    def random(self, min: int = -10, max: int = 10) -> Any:
        """
        This is a virtual method for derived classes to generate a random element from current field with degree 1
        e.g. if self is Rn hen Rn._generate_one() will return an elemnt from R1
        the generation of a full vector is with 'random' function
        """
        raise NotImplementedError("This is a virtual method")

    # def random(self, min: float = -10, max: float = 10) -> Vector:
    #     """
    #     will generate a random vector from this field
    #     """
    #     self._generate_one(min, max)
    #     return Vector .Vector([
    #         self._generate_one(min, max)
    #         for _ in range(self._degree)
    #     ], self)


class RationalField(Field):
    def __init__(self, degree=1, modulu=1):
        super().__init__(Fields.Q, 0, 1, degree, modulu)

    def random(self, min: int = -10, max: int = 10) -> float:
        if min == max:
            raise ValueError(
                "if 'min'=='max' you shouldnt use this function becuase there's no use to it")
        if min > max:
            raise ValueError("'min' should be less than 'max'")
        f = random.randint
        sign = 1 if f(0, 1) == 1 else -1
        nominator = f(min, max)
        denominator = f(min, max)
        while(denominator == 0):
            denominator = f(min, max)
        return sign*nominator/denominator

    # def __contains__(self, obj) -> bool:
    #     """
    #     NOT IMPLEMENTED
    #     Due to how numbers are stored in python all fractional numbers are rational so this function is irrelevant
    #     """
    #     raise NotImplementedError(
    #         "Due to how numbers are stored in python all fractional numbers are rational so this function is irrelevant")


DefaultRationalField = RationalField()


class RealField(Field):
    def __init__(self, degree=1, modulu=1):
        super().__init__(Fields.R, 0, 1, degree, modulu)

    def random(self, min: int = -10, max: int = 10) -> float:
        return random.uniform(min, max)

    # def __contains__(self, obj: Union[int, float, Vector.Vector]) -> bool:
    #     if isinstance(obj, float) or isinstance(obj, int) and self._degree == 1:
    #         return True
    #     else:
    #         if not isinstance(obj, Vector.Vector):
    #             raise ValueError(
    #                 "Can't check if object is not of type 'Vector'")
    #         return obj.field == self


DefaultRealField = RealField()


class ComplexField(Field):
    def __init__(self, degree=1, modulu=1):
        super().__init__(Fields.C, Complex(0, 0),
                         Complex(1, 0), degree, modulu)

    def random(self, min: int = -10, max: int = 10) -> Complex:
        return Complex.random(min, max, random.uniform)

    # def __contains__(self, obj: Union[int, float, Complex, Vector.Vector]) -> bool:
    #     if (isinstance(obj, Complex.Complex) or isinstance(obj, float) or isinstance(obj, int)) and self._degree == 1:
    #         return True
    #     else:
    #         if not isinstance(obj, Vector.Vector):
    #             raise ValueError(
    #                 "Can't check if object is not of type 'Vector'")
    #         return obj.field == self


DefaultComplexField = ComplexField()


# class MatrixField(Field):
#     # TODO MatrixFIeld

#     def __init__(self, n: int, field: Field,) -> None:

#         pass

#     def _generate_one(self, min: int = -10, max: int = 10) -> Matrix.Matrix:
#         pass

#     # def __contains__(self, v) -> bool:
#     #     pass



import random
from ...la1 import VectorSpace, Matrix, RealField
from ...la2 import BilinearForm, StandardInnerProduct as sip


def test_standard_gram_matrix():
    n = random.randint(1, 10)
    Vr2 = VectorSpace(RealField(n))
    g = BilinearForm(lambda x, y: sip(x, y), Vr2)
    assert g.toMatrix() == Matrix.id_matrix(n)



from ...la1 import RealField, Vector, LinearMap, VectorSpace

R3 = RealField(4)
R2 = RealField(2)


def test_containment():
    def func(v, target_field):
        return Vector([v[0]+v[1], v[1]+v[2]], target_field)
    lt = LinearMap(R3, R2, func)
    v = VectorSpace(R3).random()
    assert lt(v) in VectorSpace(R2)



from __future__ import annotations
from ..la1 import LinearMap, Field, Span
from typing import Callable


class Operator(LinearMap):
    def __init__(field: Field, func: Callable[[], ]) -> None:
        super().__init__(field, field, func)

    @property
    def isOrthogonoal(self) -> bool:
        pass
    # אורתוגונלי, אורתונורמלי. נורמלי. צמוד לעצמו

    def is_invariant_to(span: Span) -> bool:
        pass



from ...la1 import Matrix, RealField, VectorSpace, Vector

COUNT = 100
N = 50
V = VectorSpace(RealField(N))


def test_from_vector():
    for _ in range(COUNT):
        v = V.random()
        m = Matrix.fromVector(v)
        for i in range(v.length):
            assert v[i] == m[i][0]


def test_from_vectors():
    for _ in range(int(COUNT/10)+1):
        vecs = [V.random() for _ in range(10)]
        m = Matrix.fromVectors(vecs)
        for i in range(len(vecs[0])):
            for j in range(len(vecs)):
                assert vecs[j][i] == m[i][j]


def test_determinant():
    assert Matrix([[1, 1], [1, 1]]).determinant == 0


def test_id():
    assert Matrix.id_matrix(2) == Matrix([[1, 0], [0, 1]])


def test_guassian_elimination():
    assert Matrix([[1, 1], [1, 1]]).guassian_elimination(
    )[0] == Matrix([[1, 1], [0, 0]])
    assert Matrix.id_matrix(5).guassian_elimination(
    )[0] == Matrix.id_matrix(5)


def test_solve():
    assert Matrix([[1, 3], [2, 5]]).solve(Vector([1, 2])) == Vector([1, 0])
    assert Matrix([[1, 1], [1, 1]]).solve(Vector([1, 2])) == None
    assert Matrix.id_matrix(2).solve(Vector([0, 0])) == Vector([0, 0])
    # FIXME
    # assert Matrix([[1, 1], [1, 1]]).solve() == Span([Vector([0, 1])])


# def test_kernel():
#     from Span import Span
#     from Vector import Vector
#     assert Matrix([[1, 0], [0, 0]]).kernel == Span([Vector([0, 1])])


# test_kernel()



# from ...la1 import Matrix, MatrixOperation, MatrixOperationType


# def test_row_multiplication():
#     assert (Matrix([[5, 0], [0, 1]]), None) == MatrixOperation(MatrixOperationType.ROW_MULTIPLICATION,
#                                                                0, 5)(Matrix.id_matrix(2))


# def test_row_switching():
#     assert (Matrix([[0, 1], [1, 0]]), None) == MatrixOperation(
#         MatrixOperationType.ROW_SWITCHING, 0, 1)(Matrix.id_matrix(2))


# def test_row_addition():
#     assert (Matrix([[1, 1], [0, 1]]), None) == MatrixOperation(
#         MatrixOperationType.ROW_ADDITION, 0, 1)(Matrix.id_matrix(2))



import random
from ...utils import *
from ...la1 import Complex as comp, Vector, RealField, LinearTransformation, VectorSpace


def test_almost_equal():
    assert almost_equal(comp(1, -5.551115123125783e-17),
                        comp(1, 0), comp(1, -5.551115123125783e-17))
    a, b, c = [-0.6666666666666666, -0.3333333333333333, 1.0]
    assert almost_equal((a+b)+c, a+(b+c))
    assert almost_equal(round(0.9999999999999999), 1)


def test_bracketify():
    assert bracketify("()") == [("(", 0), (")", 1)]


def test_insert_multiplication():
    func = insert_multiplication
    assert func("(()())(())") == '(()*())*(())'
    assert func(
        "(())(()()()(()))(())") == '(())*(()*()*()*(()))*(())'
    assert func("()()+()()") == '()*()+()*()'


def test_open_power():
    func = open_power
    assert func("(x+1)^2") == "(x+1)(x+1)"
    assert func("((x+1)^2)^2") == "((x+1)(x+1))((x+1)(x+1))"
    assert func("(x)^10") == "(x)(x)(x)(x)(x)(x)(x)(x)(x)(x)"


def test_split_not_between_brackets():
    func = split_not_between_brackets
    assert func("(()())*(()())", "*") == ["()*()", "()*()"]
    assert func("()()()()", "*") == ["", "", "", ""]
    assert func("(x+1)(x+1)", "*") == ["x+1", "x+1"]
    assert func("(x+1)(x+1)", "+") == ["x+1", "x+1"]
    assert func("(x+1)(x+1)", "-") == ["x+1", "x+1"]
    assert func("(x+1)(x+1)", "/") == ["x+1", "x+1"]
    assert func("((x)+1)(x+1)", "*") == ["(x)+1", "x+1"]


def test_composit_functions():
    def func(v, target):
        return Vector([v[1], v[0]], target)
    R2 = RealField(2)
    lt = LinearTransformation(R2, R2, func)
    v = VectorSpace(R2).random()
    assert lt(lt(v)) == v, 1
    assert (lt**2)(v) == v, 2
    assert (lt**4)(v) == (lt**2)(v), 3
    num = random.randint(1, 10)
    lt2 = num*lt
    assert lt2(v) == Vector([v[1]*num, v[0]*num], R2), 4
    assert (lt2**2)(v).almost_equal(Vector([v[0]*num**2, v[1]*num**2], R2))

    def func(v, target):
        return Vector([0, v[0]], target)
    lt = LinearTransformation(R2, R2, func)
    v = VectorSpace(R2).random()
    assert (num*lt**2)(v) == Vector([0, 0], R2), 6



from __future__ import annotations
from typing import Tuple, Union
import functools
from .Calculable import Calculable
from ...la1 import Vector, Complex, Matrix, LinearMap
from ...utils import *


class PolynomialSimple(Calculable):
    @staticmethod
    def fromString(input: str, var: str = "x") -> PolynomialSimple:
        if var not in input:
            for c in input:
                if not c.isdigit():
                    raise ValueError("")
            return PolynomialSimple([float(input)], [0])
            raise ValueError("Variable not found in string")
        # --------helper functions----------

        def splitter(lst: list[str], symbol: str, excluders: list[str] = list(bracket_pairs.keys())) -> list[str]:
            res = []
            for string in lst:
                res.extend(string.strip().split(symbol))
            return res

        def stacker(string: str, lst: list[str]) -> list[str]:
            res = []
            for c in string:
                if c in lst:
                    res.append(c)
            return res

        def sub_to_poly(sub) -> PolynomialSimple:
            def pow_to_poly(subb: str) -> PolynomialSimple:
                individual_chars = splitter([subb], "^")
                if len(individual_chars) == 2:
                    bias = 1

                    def extract_nums(indiv) -> Tuple[float, float]:
                        try:
                            power = float(indiv[bias % 2])
                            prefix = 1
                            if indiv[(bias+1) % 2] != var:
                                prefix = float(
                                    indiv[(bias+1) % 2].replace(var, ""))
                            return prefix, power
                        except Exception as e:
                            raise e

                    prefix, power = 1, 0
                    try:
                        prefix, power = extract_nums(individual_chars)
                    except ValueError as e:
                        bias += 1
                        prefix, power = extract_nums(individual_chars)
                    return PolynomialSimple([prefix], [power])

                elif len(individual_chars) == 1:
                    text = individual_chars[0]
                    if text == var:
                        return PolynomialSimple([1], [1])
                    prefix = float(text.replace(var, ""))
                    return PolynomialSimple([prefix], [1 if var in text else 0])
                assert False, "didnt think about that"

            multiplication_stack = stacker(sub, ["*", "/"])
            subs2 = splitter([sub], "*")
            subs2 = splitter(subs2, "/")
            res = pow_to_poly(subs2[0])
            for i, s in enumerate(subs2[1:]):
                if multiplication_stack[i] == "*":
                    res *= pow_to_poly(s)
                elif multiplication_stack[i] == "/":
                    res /= pow_to_poly(s)

            return res

        def sign(op): return 1 if op == "+" else -1

        # handle if input has brackets
        if any(bracket in input for bracket in ["(", ")", "[", "]", "{", "}"]):
            if not validate_brackets(input):
                raise ValueError("invalid brackets")
            input = open_power(input)
            sub_inputs = split_not_between_brackets(input, "*")
            res = PolynomialSimple.fromString(sub_inputs[0], var)
            for sub_input in sub_inputs[1:]:
                res *= PolynomialSimple.fromString(sub_input, var)
            return res
        else:
            # split with addition
            addition_stack = stacker(input, ["+", "-"])
            subs = splitter([input], "+")
            subs = splitter(subs, "-")
            subs = [sub for sub in subs if len(sub) > 0]
            if len(addition_stack) == len(subs)-1:
                addition_stack.insert(0, "+")

            # foreach sub expression create a Poly with inner splitting by multiplication and add it acording to current addition operator
            res = sign(addition_stack[0])*sub_to_poly(subs[0])
            for i, sub in enumerate(subs[1:]):
                res += sign(addition_stack[i])*sub_to_poly(sub)
            return res

    def __init__(self, prefixes: list, powers: list) -> None:
        if len(prefixes) != len(powers):
            if len(prefixes) == 1 and len(powers) == 0 and prefixes[0] == 0:
                self.powers = []
                self.prefixes = []
                return
            raise ValueError("Prefixes and powers must be of same length")

        tuples = [(prefixes[i], powers[i]) for i in range(len(powers))]
        POW = 1
        PRE = 0

        def comparer(v1: Tuple, v2: Tuple) -> int:
            # TODO what if one of them is complex? __lt__ need to be implemented, but how?
            bias = 1
            if v1[POW] < v2[POW]:
                return -bias
            elif v1[POW] > v2[POW]:
                return bias
            else:
                if v1[PRE] < v2[PRE]:
                    return -bias
                elif v1[PRE] > v2[PRE]:
                    return bias
                else:
                    return 0

        tuples.sort(key=functools.cmp_to_key(comparer), reverse=True)
        powers = [tuples[i][POW]
                  for i in range(len(tuples)) if tuples[i][PRE] != 0]
        prefixes = [tuples[i][PRE]
                    for i in range(len(tuples)) if tuples[i][PRE] != 0]
        self.prefixes = []
        self.powers = []
        i = 0
        while i < len(powers):
            p = powers[i]
            count = powers.count(p)
            if count > 1:
                r = range(i, i+count)
                sum = 0
                for j in r:
                    sum += prefixes[j]
                    i += 1
                i -= 1
                self.prefixes.append(sum)
                self.powers.append(p)
            else:
                self.prefixes.append(prefixes[i])
                self.powers.append(p)
            i += 1
        self.powers = [self.powers[i]
                       for i in range(len(self.powers)) if self.prefixes[i] != 0]
        self.prefixes = [self.prefixes[i]
                         for i in range(len(self.prefixes)) if self.prefixes[i] != 0]

    @property
    def roots(self) -> Vector:
        # TODO calculation of roots if possible
        pass

    @property
    def degree(self) -> float:
        return self.powers[0] if len(self.powers) > 0 else 0

    def __str__(self) -> str:
        def one_to_str(i: int) -> str:
            power = self.powers[i]
            if power == int(power):
                power = int(power)
            prefix = self.prefixes[i]
            if prefix == int(prefix):
                prefix = int(prefix)
            res = ""

            if prefix == 0:
                return ""
            elif abs(prefix) == 1:
                if power == 0:
                    return "1"
                if prefix == 1:
                    return "X" if power == 1 else "X^" + str(power)
                else:
                    return "-X" if power == 1 else "-X^" + str(power)
            else:
                res += str(prefix)

            if power == 0:
                return res
            elif power == 1:
                return res+"X"
            return f"{str(prefix)}X^{str(power)}"
        return " + ".join([one_to_str(i) for i in range(len(self))])

    def __add__(self, other) -> PolynomialSimple:
        if isoneof(other, [int, float, Complex]):
            if other == 0:
                return PolynomialSimple(self.prefixes, self.powers)
            elif 0 in self.powers:
                index = self.powers.index(0)
                if self.prefixes[index] + other == 0:
                    return PolynomialSimple(self.prefixes[:index]+self.prefixes[index+1:], self.powers[:index]+self.powers[index+1:])
                new_prefixes = self.prefixes[:index] + \
                    [self.prefixes[index] + other] + self.prefixes[index+1:]
                return PolynomialSimple(new_prefixes, self.powers)
            return PolynomialSimple(self.prefixes+[other], self.powers+[0])
        elif isinstance(other, PolynomialSimple):
            return PolynomialSimple(self.prefixes+other.prefixes, self.powers+other.powers)
        raise NotImplementedError(
            f"Polynomial addition not implemented for {type(other)}")

    def __radd__(self, other) -> PolynomialSimple:
        return self.__add__(other)

    def __sub__(self, other) -> PolynomialSimple:
        return self.__add__(-other)

    def __rsub__(self, other) -> PolynomialSimple:
        return other + (-self)

    def __neg__(self) -> PolynomialSimple:
        return self*(-1)

    def __mul__(self, other) -> Union[PolynomialSimple, Matrix]:
        if isoneof(other, [int, float, Complex]):
            return PolynomialSimple([other*pre for pre in self.prefixes], self.powers)
        elif isinstance(other, PolynomialSimple):
            new_prefixes, new_powers = [], []
            for i in range(len(other)):
                for j in range(len(self)):
                    new_prefixes.append(self.prefixes[j]*other.prefixes[i])
                    new_powers.append(self.powers[j]+other.powers[i])
            return PolynomialSimple(new_prefixes, new_powers)
        elif isinstance(other, Matrix):
            return other*self
        raise NotImplementedError(
            f"Polynomial multiplication not implemented for {type(other)}")

    def __rmul__(self, other) -> PolynomialSimple:
        return self.__mul__(other)

    def __truediv__(self, other) -> Tuple[PolynomialSimple, PolynomialSimple]:
        if self == other:
            return PolynomialSimple([1], [0]), 0
        if isoneof(other, [int, float, Complex]):
            return self*(1/other), 0
        elif isinstance(other, PolynomialSimple):
            if self.degree >= other.degree and len(self) > len(other):
                quotient, remainder = PolynomialSimple([0], [0]), None
                current = self
                prefix, power = 1, 1
                while(current.degree >= other.degree):
                    prefix = current.prefixes[0]/other.prefixes[0]
                    power = current.prefixes[0] - other.powers[0]
                    quotient += PolynomialSimple([prefix], [power])
                    subtructor = PolynomialSimple(
                        [quotient.prefixes[-1]*v for v in other.prefixes], [quotient.prefixes[-1]*v for v in other.powers])
                    current -= subtructor
                remainder, _ = current/other
                return quotient, remainder
            else:
                if self.degree == 0:
                    return self.prefixes[0]/other, 0
                return None, 0
        # TODO need to add?
        raise NotImplementedError("Polynomial division not implemented")

    def __rtruediv__(self, other) -> PolynomialSimple:
        if isoneof(other, [int, float, Complex]):
            pass
        elif isinstance(other, PolynomialSimple):
            pass

    def __pow__(self, pow) -> PolynomialSimple:
        if isinstance(pow, int):
            if pow == 0:
                return PolynomialSimple([1], [0])
            res = self
            for _ in range(abs(pow)-1):
                res *= self
            if pow < 0:
                return 1/res
            return res

    def __eq__(self, other: PolynomialSimple) -> bool:
        if isoneof(other, [int, float, Complex]):
            if other == 0:
                return len(self) == 0
            if len(self) != 1:
                return False
            return self.powers[0] == 0 and self.prefixes[0] == other
        if isinstance(other, PolynomialSimple):
            if len(self) != len(other):
                return False
            for i in range(len(self)):
                if self.prefixes[i] != other.prefixes[i] or self.powers[i] != other.powers[i]:
                    return False
            return True
        raise NotImplementedError(
            f"Polynomial equality not implemented for {type(other)}")

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __call__(self, v):
        from .PolynomialFraction import PolynomialFraction
        if not isoneof(v, [int, float, Complex, PolynomialSimple, Matrix, LinearMap, PolynomialFraction]):
            # TODO implement __call_ for matrix,operator,vector,etc
            raise NotImplementedError(
                "Polynomial __call__ not implemented for " + str(type(v)))
        res = self.prefixes[0]*v**self.powers[0]
        for i in range(1, len(self)):
            res += self.prefixes[i]*v**self.powers[i]
        return res

    def __len__(self) -> int:
        return len(self.powers)

    def solve(self):
        pass

    def gcd_with(self, other: PolynomialSimple) -> PolynomialSimple:
        # TODO implement gcd calculation
        pass

    # TODO implement other operators



import random
from typing import Callable, Union
from ..la1 import Complex, Vector


class InnerProduct:

    @staticmethod
    def isInnerProduct(func: Callable[[Vector, Vector], float], generator_func: Callable[[bool], Vector]) -> bool:
        MIN_VAL = -100
        MAX_VAL = 100
        REPETITIONS = 1000
        THRESHOLD = 0.000000001

        def check_linearity(func: Callable[[Vector, Vector], float], generator_func: Callable[[bool], Vector]) -> bool:
            for _ in range(REPETITIONS):
                a: Vector = generator_func()
                b: Vector = generator_func()
                c: Vector = generator_func()
                s1 = random.randint(MIN_VAL, MAX_VAL)
                s2 = random.randint(MIN_VAL, MAX_VAL)
                original = func(a, s1*b+s2*c)
                decomp = s1*func(a, b)+s2*func(a, c)
                if original - decomp > THRESHOLD:
                    return False
            return True

        def check_symmetry(func: Callable[[Vector, Vector], float], generator_func: Callable[[bool], Vector]) -> bool:
            for _ in range(REPETITIONS):
                a: Vector = generator_func()
                b: Vector = generator_func()
                # TODO implemnt symmetry check for complex
                original = func(a, b)
                adj = func(b, a)
                if original-adj > THRESHOLD:
                    return False
            return True

        def check_norm(func: Callable[[Vector, Vector], float], generator_func: Callable[[bool], Vector]) -> bool:
            for _ in range(REPETITIONS):
                v: Vector = generator_func()
                if not func(v, v) > 0:
                    return False
            v = generator_func(True)
            if func(v, v) != 0:
                return False
            return True

        return all[check_linearity(func, generator_func), check_symmetry(func, generator_func), check_norm(func, generator_func)]

    def __init__(self, func: Callable[[Vector, Vector], float]) -> None:
        if func is None:
            raise ValueError("func is None")
        self._func = func

    def __call__(self, v1, v2) -> Union[float, Complex]:
        return self._func(v1, v2)


# def __StandardInnerProduct(a: Vector, b: Vector) -> float:
#     if not isinstance(a, Vector):
#         raise TypeError("a must be a Vector")
#     if not isinstance(b, Vector):
#         raise TypeError("b must be a Vector")
#     if a.length != b.length:
#         raise ValueError("Vectors must have the same length")
#     return sum([a[i] * b[i] for i in range(a.length)])

StandardInnerProduct = InnerProduct(
    lambda a, b: sum([a[i] * b[i] for i in range(a.length)]))



from __future__ import annotations
from typing import Union
from ..utils import areinstances, check_foreach
from ..la1 import VectorSpace, Field, Complex, Vector
from .InnerProduct import InnerProduct


class InnerProductSpace(VectorSpace):
    def __init__(self, field: Field, inner_product: InnerProduct) -> None:
        super().__init__(field)
        self.inner_product = inner_product

    def apply(self, v: Vector, u: Vector) -> Union[int, float, Complex]:
        return self(v, u)

    def inner_product(self, v: Vector, u: Vector) -> Union[int, float, Complex]:
        return self(v, u)

    def __call__(self, v: Vector, u: Vector) -> Union[int, float, Complex]:
        if not areinstances([v, u], Vector):
            raise TypeError("v and u must be of type Vector")
        return self.inner_product(v, u)

    def arePerpendicular(self, v: Vector, u: Vector) -> bool:
        if not areinstances([v, u], Vector):
            raise TypeError("v and u must be of type Vector")
        if not check_foreach([v, u], lambda x: x in self):
            raise ValueError("v and u must be in the VectorSpace")
        return self(v, u) == 0



from __future__ import annotations
from typing import Any
from .Vector import Vector
from .Field import Field
from ..utils import are_operators_implemnted


class Span:
    @staticmethod
    def spanField(field: Field) -> Span:
        """
        will return the standard span over F^n e.g : [e1,e2,...,en]
        """
        vecs = []
        for i in range(field._degree):
            v = Vector.fromSize(field._degree, field._zero)
            v[i] = field._one
            vecs.append(v)
        return Span(vecs)

    def __init__(self, base: list[Vector] = []) -> None:
        """
        Initialize a Span object.
        :param base: A list of objects to be used as the base of the span.
        :param validate: If True, the span will be validated as a vector space
        """
        if base != []:
            example_item = base[0]
            T = type(example_item)
            for val in base:
                if not isinstance(val, T):
                    raise TypeError(
                        "All elements of the base must be of the same type")

            if not hasattr(example_item, "field"):
                raise TypeError(
                    "All objects must have a field attribute which is an Instance of class Field")
            # if all vectors in base are of the same field add then otherwise throw an error
            self.field = example_item.field
            for vector in base:
                if vector.field != self.field:
                    raise ValueError(
                        "Span can only be created from vectors of the same field")
        self.vectors = base
        if not are_operators_implemnted(type(self.vectors[0])):
            raise AttributeError(
                "Not all required operators are implemented for the class of the objects")

    @property
    def dim(self) -> int:
        return len(self)

    @property
    def basis(self) -> Span:
        pass

    @property
    def has_lineary_dependency(self) -> bool:
        pass

    @property
    def is_orthogonal(self):
        pass

    @property
    def is_orthonormal(self):
        pass

    def __str__(self) -> str:
        result = ""
        for vector in self.vectors:
            result += str(vector)+"\n"
        return result

    def __add__(self, other: Span) -> Span:
        if not isinstance(other, Span):
            raise TypeError("Span can only be added to another Span")
        if len(self.vectors) != len(other.vectors):
            raise ValueError("Spans must have the same length")
        return Span([self.vectors[i] + other.vectors[i] for i in range(len(self.vectors))])

    def __getitem__(self, index: int) -> Vector:
        return self.vectors[index]

    def __iter__(self):
        return iter(self.vectors)

    def __len__(self) -> int:
        return len(self.vectors)
    # def validate(self) -> bool:
    #     from InnerProduct import StandardInnerProduct
    #     for i in range(len(self.vectors)-1):
    #         for j in range(i+1, len(self.vectors)):
    #             if StandardInnerProduct(self.vectors[i], self.vectors[j]) != 0:
    #                 return False
    #     return True

    def __contains__(self, vector: Vector) -> bool:
        """
        operator 'in' wil specify wheter an element (a Vector) exsists in the vectors lsit of the span
        """
        if not isinstance(vector, Vector):
            raise TypeError(
                "can only check containment of objects of type 'Vector'")
        for v in self.vectors:
            if vector == v:
                return True
        return False

    def contains(self, vector: Vector) -> bool:
        """
        will return true if there is a linear combination of self's vectors that creates 'vector'
        """
        if not isinstance(vector, Vector):
            raise TypeError(
                "can only check containment of objects of type 'Vector'")
        from Matrix import Matrix
        Matrix.fromSpan(self, vector)
        return False

    def append(self, vec: Vector) -> None:
        self.vectors.append(vec)

    def toOrthonormal(self) -> Span:
        result = [self[0].toOrthonormal()]
        from ..la2 import StandardInnerProduct as sip
        for i in range(1, len(self.vectors)):
            current = self[i]
            curr_tag = Vector([0 for _ in range(self[0].length)])
            for prev in result:
                curr_tag = curr_tag+sip(prev, current) * prev
            current = current-curr_tag
            result.append(current.toOrthonormal())
        return Span(result)

    def projection_of(self, v: Vector) -> Vector:
        """
        returns the vector projection of vector v on the span (=self)
        """
        res: Vector = Vector.fromSize(len(v), 0)
        for w in self.toOrthonormal():
            res += v.projection_onto(w)
        return res

    def random(self, min: Any = -10, max: Any = 10) -> Vector:
        return self.field.random(min, max)

    def is_spanning(self, field: Field) -> bool:
        # TODO implement this
        raise NotImplementedError("")



from __future__ import annotations
from typing import Callable
from .Field import Field, RealField
from .Vector import Vector


class VectorSpace:

    def __init__(self, field: Field) -> None:
        self.field = field

    def __str__(self) -> str:
        pass

    def __eq__(self, other: VectorSpace) -> bool:
        if not isinstance(other, VectorSpace):
            raise TypeError("other must be an instance of class VectorSpace")
        return self.field == other.field

    def __ne__(self, other: VectorSpace) -> bool:
        return not self.__eq__(other)

    def __contains__(self, value: Vector) -> bool:
        if not isinstance(value, Vector):
            raise TypeError("value must be an instance of class Vector")
        if value.field == self.field:
            return True
        return False

    def random(self, min: int = -10, max: int = 10) -> Vector:
        return Vector([self.field.random(min, max) for _ in range(self.field._degree)], self.field)

    def standard_basis(self) -> list[Vector]:
        n = self.field._degree
        empty = [0 for _ in range(n)]
        return [Vector(empty[:i]+[1]+empty[i+1:]) for i in range(n)]


# class PolynomialVectorSpace(VectorSpace):
#     pass

# TODO is oclidian, is hermitian



from __future__ import annotations
from typing import Union
import random
from ..utils import isoneof, almost_equal


class Complex:
    def __init__(self, real: float, imag: float) -> None:
        self.real = real
        self.imag = imag

    def __str__(self) -> str:
        # get final string for real value
        real = self.real
        if almost_equal(round(self.real), self.real):
            real = str(round(self.real))
        else:
            real = f"{real:.2f}"
        # get final string for imag value
        imag = self.imag
        if almost_equal(round(self.imag), self.imag):
            imag = str(round(self.imag))
        else:
            imag = f"{imag:.2f}"
        # return string
        if real == imag == "0":
            return "0"
        elif real != "0" == imag:
            return real
        elif real == "0" != imag:
            return imag+"i"
        return f'{real} + {imag}i'

    def __add__(self, other: Union[float, Complex]) -> Complex:
        if isinstance(other, float):
            return Complex(self.real + other, self.imag)
        else:
            return Complex(self.real + other.real, self.imag + other.imag)

    def __radd__(self, other: Union[float, Complex]) -> Complex:
        return self.__add__(other)

    def __mul__(self, other: Union[float, Complex]) -> Complex:
        if isinstance(other, float):
            return Complex(self.real * other, self.imag * other)
        else:
            return Complex(self.real * other.real - self.imag * other.imag,
                           self.real * other.imag + self.imag * other.real)

    def __rmul__(self, other: Union[float, Complex]) -> Complex:
        return self.__mul__(other)

    def __neg__(self) -> Complex:
        return Complex(-self.real, -self.imag)

    def __sub__(self, other: Union[float, Complex]) -> Complex:
        return self + (-other)

    def __rsub__(self, other: Union[float, Complex]) -> Complex:
        # other-self==-(self-other)
        return -self.__sub__(other)

    def __abs__(self) -> float:
        return (self.real ** 2 + self.imag ** 2) ** 0.5

    def __eq__(self, other: Complex) -> bool:
        if not isoneof(other, [int, float, Complex]):
            raise TypeError(f"cannot compare equality to type {type(other)}")
        if not isinstance(other, Complex):
            other = Complex(other, 0)
        return almost_equal(self.real, other.real) and almost_equal(self.imag, other.imag)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __truediv__(self, other: Union[int, float, Complex]) -> Complex:
        if not isinstance(other, Complex) and not isinstance(other, float) and not isinstance(other, int):
            raise TypeError(
                "can only divide complex by [int, float , complex]")
        if other == 0:
            raise ZeroDivisionError("Cannot divide by zero")

        if not isinstance(other, Complex):
            other = Complex(other, 0)

        nominator = Complex(self.real, self.imag)*other.conjugate
        denominator = other*other.conjugate
        return Complex(nominator.real/denominator.real, nominator.imag/denominator.real)

    def __rtruediv__(self, other) -> Complex:
        if not isinstance(other, Complex) and not isinstance(other, float) and not isinstance(other, int):
            raise TypeError(
                "can only divide [int, float , complex] by complex")
        if not isinstance(other, Complex):
            other = Complex(other, 0)
        nominator = other*self.conjugate
        denominator = self*self.conjugate
        return Complex(nominator.real/denominator.real, nominator.imag/denominator.real)

    def __pow__(self, p):
        # TODO: implement Complex**p fully
        if not isoneof(p, [int]):
            raise NotImplementedError(
                "Complex.__pow__ only implemented for int right now")

        def fix_negativ(v) -> Complex:
            if p < 0:
                return 1/v
            return v

        if p == 0:
            return Complex(1, 0)
        elif p == 1:
            return fix_negativ(self)
        else:
            res = self
            for _ in range(abs(p)-1):
                res *= self
            return fix_negativ(res)

    def __rpow__(self, other):
        # TODO: implement base**Complex
        raise NotImplementedError("This is not implemented yet")

    @ property
    def conjugate(self):
        return Complex(self.real, -self.imag)

    @ property
    def norm(self):
        return (self * self.conjugate).real

    @ staticmethod
    def random(min_val: float = -10, max_val: float = 10, value_func=random.randint) -> Complex:
        return Complex(value_func(min_val, max_val), value_func(min_val, max_val))



from __future__ import annotations


class Calculable:
    @staticmethod
    def fromString(s: str) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __init__(self):
        try:
            self.__add__
            self.__radd__
            self.__mul__
            self.__rmul__
            self.__neg__
            self.__sub__
            self.__rsub__
            self.__truediv__
            self.__rtruediv__
            self.__pow__
            self.__str__
            self.__eq__
            self.__ne__
        except NotImplementedError:
            raise NotImplementedError(
                "abstract class - operators are not implemented")

    def __add__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __radd__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __mul__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __rmul__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __neg__(self) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __sub__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __rsub__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __truediv__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __rtruediv__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __pow__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __eq__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __ne__(self, other) -> Calculable:
        raise NotImplementedError("This is a virtual method")

    def __call__(self, value):
        raise NotImplementedError("This is a virtual method")

    def __str__(self) -> str:
        raise NotImplementedError("This is a virtual method")



# import Complex

import math
from typing import Tuple


def are_operators_implemnted(T) -> bool:
    try:
        T.__add__
        T.__radd__
        T.__sub__
        T.__rsub__
        T.__neg__
        T.__mul__
        T.__rmul__
        T.__truediv__
        T.__rtruediv__
        T.__eq__
        T.__ne__
        return True
    except AttributeError:
        return False


def isoneof(v, lst: list) -> bool:
    for T in lst:
        if isinstance(v, T):
            return True
    return False


def alloneof(value: list, types: list):
    for v in value:
        if not isoneof(v, types):
            return False
    return True


def almost_equal(*args):
    THRESHOLD = 0.000000000001
    # def compare_two(a, b) -> bool:
    #     # if b == 0:
    #     #     try:
    #     #         return a < THRESHOLD
    #     #     except:
    #     #         return a == 0
    #     # if a == 0:
    #     #     try:
    #     #         return b < THRESHOLD
    #     #     except:
    #     #         return b == 0
    #     if abs(a) > 100000 and abs(b) > 100000:
    #         return abs(1-a/b) < THRESHOLD
    #     return abs(a-b) < THRESHOLD
    # return all([compare_two(args[0], args[i]) for i in range(1, len(args))])

    def wrapper(a, b):
        if alloneof([a, b], [int, float]):
            return math.isclose(a, b, abs_tol=THRESHOLD)
        else:  # they are Complex.Complex
            try:
                return math.isclose(a.real, b.real, abs_tol=THRESHOLD) and math.isclose(a.imag, b.imag, abs_tol=THRESHOLD)
            except Exception as e:
                assert False, "shouldnt be here"
    return all([wrapper(args[0], args[i]) for i in range(1, len(args))])


def areinstances(arr: list, T):
    return check_foreach(arr, lambda v: isinstance(v, T))


def check_foreach(arr: list, condition) -> bool:
    for v in arr:
        if not condition(v):
            return False
    return True


bracket_pairs = {
    "(": ")",
    "[": "]",
    "{": "}",
    ")": "(",
    "]": "[",
    "}": "{"
}


def validate_brackets(s: str) -> bool:
    if len(s) == 0:
        return True
    brackets = ["(", ")", "[", "]", "{", "}"]
    stack = []
    for c in s:
        if c in brackets:
            if c in ["(", "[", "{"]:
                stack.append(c)
            else:
                if len(stack) == 0:
                    return False
                if stack[-1] != brackets[brackets.index(c)-1]:
                    return False
                stack.pop()
    return len(stack) == 0


open_brackets = ["(", "[", "{"]
close_brackets = [")", "]", "}"]


def bracketify(input) -> list[Tuple[str, int]]:
    res = []
    for i, c in enumerate(input):
        if c in open_brackets+close_brackets:
            res.append((c, i))
    return res


def insert_multiplication(input: str) -> str:
    brackets = bracketify(input)
    stack = []
    # recent = None
    change = 0
    for i in range(len(brackets)-1):
        if brackets[i][0] in open_brackets:
            stack.append(brackets[i])
        elif brackets[i][0] in close_brackets:
            if stack[-1][0] == bracket_pairs[brackets[i][0]]:
                # recent = stack.pop()
                if brackets[i+1][0] in open_brackets and brackets[i+1][1] == brackets[i][1] + 1:
                    input = f"{input[:brackets[i][1]+1+change]}*{input[brackets[i][1]+1+change:]}"
                    change += 1
    return input


def open_power(input: str) -> str:
    stack = []
    i = 0
    change = 0
    while i < len(input)-1:
        c = input[i]
        if c in open_brackets:
            stack.append((c, i))
        else:
            if c in close_brackets:
                if stack[-1][0] == bracket_pairs[c]:
                    recent = stack.pop()
                    if input[i+1] == "^":
                        try:
                            # extract power end index
                            j = i+2
                            while j < len(input):
                                c2 = input[j]
                                if not c2.isdigit():
                                    break
                                j += 1
                            # parse power
                            power = int(input[i+2:j])
                            # amount of digits in the power
                            size = len(str(power))
                            # the text to insert 'power' times
                            text_to_insert = input[recent[1]:i+1]

                            # insertion
                            change = 0
                            for _ in range(power-1):
                                input = f"{input[:i+1+change]}{text_to_insert}{input[i+change+2+size:]}"
                                change += len(text_to_insert)
                            i += change
                        except Exception as e:
                            pass

        i += 1
    return input


def split_not_between_brackets(input: str, symbol: str) -> list[str]:
    input = insert_multiplication(input)
    indecies = [0]
    stack = []
    for i, c in enumerate(input):
        if c in open_brackets:
            stack.append(c)
        elif c in close_brackets:
            if stack[-1] == bracket_pairs[c]:
                stack.pop()
            if len(stack) == 0:
                indecies.append(i)
    res = []
    for i in range(len(indecies)-1):
        res.append(input[indecies[i]+1+2*i:indecies[i+1]])
    return res


def composite_function(f, g):
    return lambda *args: f(g(*args), *args[1:])



from __future__ import annotations
from ..utils import almost_equal, isoneof
from typing import Union, Any
from .Field import Field, RealField, DefaultRealField
from .Complex import Complex


class Vector:

    @staticmethod
    def random(min: float = -10, max: float = 10, degree: int = 10,  def_value=None, f: Field = DefaultRealField) -> Vector:
        return Vector([f.random(min, max) if def_value is None else def_value for _ in range(degree)])

    @staticmethod
    def fromSize(size: int, default_value: Any = 0) -> Vector:
        return Vector([default_value for _ in range(size)])

    def __init__(self, values: list[Any], field: Field = None) -> None:
        self.__values = values
        # TODO add default field detection
        self.field = RealField(len(values)) if not field else field

    @property
    def length(self):
        return len(self.__values)

    @property
    def conjugate(self) -> Vector:
        return [v.conjugate if isinstance(v, Complex) else v for v in self]

    @property
    def has_no_zero(self) -> bool:
        for v in self:
            if v != 0:
                return True
        return False

    def __str__(self) -> str:
        result = "["
        for v in self:
            result += str(v) + ", "
        return result[:-2]+"]"

    def __add__(self, other: Vector) -> Vector:
        if not isinstance(other, Vector):
            raise TypeError("Vector can only be added to another Vector")
        if self.field != other.field:
            raise ValueError("Vectors must have the same field")
        if len(self.__values) != len(other.__values):
            raise ValueError("Vectors must have the same length")
        return Vector([self[i] + other[i] for i in range(len(self))], self.field)

    def __radd__(self, other) -> Vector:
        return self.__add__(other)

    def __sub__(self, other: Vector) -> Vector:
        if not isinstance(other, Vector):
            raise TypeError(
                "Vector can only be subtracted from another Vector")
        if self.field != other.field:
            raise ValueError("Vectors must have the same field")
        if len(self.__values) != len(other.__values):
            raise ValueError("Vectors must have the same length")
        return Vector([self.__values[i] - other.__values[i] for i in range(len(self.__values))], self.field)

    def __rsub__(self, other) -> Vector:
        return (-self) + other

    def __neg__(self) -> Vector:
        return Vector([-self.__values[i] for i in range(len(self.__values))], self.field)

    def __mul__(self, num: float) -> Vector:
        return Vector([num * self.__values[i] for i in range(len(self.__values))], self.field)

    def __rmul__(self, num: float) -> Vector:
        return self.__mul__(num)

    def __truediv__(self, other) -> Vector:
        if isoneof(other, [int, float, Complex]):
            return self.__mul__(1/other)
        raise ValueError("cant divide vector")

    def __rtruediv__(self, num: float) -> Vector:
        raise ValueError("cant divide by vector")

    def __getitem__(self, index: int) -> Union[int, float, Complex]:
        return self.__values[index]

    def __iter__(self):
        return iter(self.__values)

    def __eq__(self, other: Vector) -> bool:
        if not isoneof(other, [Vector, list]):
            return False
        if isinstance(other, Vector):
            other = other.__values
        if len(self) != len(other):
            return False
        return self.__values == other

    def __ne__(self, other: Vector) -> bool:
        return not (self == other)

    def __len__(self) -> int:
        return self.length

    def almost_equal(self, other: Vector) -> bool:
        if not isinstance(other, Vector):
            raise TypeError("Vector can only be compared to another Vector")
        if not self.field == other.field:
            raise ValueError("Vectors must have the same field")
        return all([almost_equal(self[i], other[i]) for i in range(len(self))])

    def set(self, index, value) -> None:
        self.__values[index] = value

    def norm(self) -> float:
        return sum([x ** 2 for x in self]) ** 0.5

    def dot(self, other: Vector) -> Vector:
        if not isinstance(other, Vector):
            raise TypeError("Vector can only be multiplied by another Vector")
        if self.field != other.field:
            raise ValueError("Vectors must have the same field")
        if len(self.__values) != len(other.__values):
            raise ValueError("Vectors must have the same length")
        return Vector([self.__values[i] * other.__values[i] for i in range(len(self.__values))])

    def toOrthonormal(self) -> Vector:
        # TODO change name to Normalize?
        return self / self.norm()

    def projection_onto(self, value) -> Vector:
        """
        return the projection of self onto value which can be another vector or a Span
        """
        from .Span import Span
        if not isinstance(value, Vector) and not isinstance(value, Span):
            raise TypeError("v must be of type Vector or Span")
        is_span = isinstance(value, Span)
        if not is_span and value.length != self.length:
            raise ValueError("value must have the same length as self")
        if is_span and value[0].length != self.length:
            raise ValueError(
                "the span's vectors must have the same length as self")
        from ..la2.InnerProduct import StandardInnerProduct as sip
        if not is_span:
            return sip(self, value)/sip(value, value)*value
        else:
            return value.projection_of(self)

    def copy(self) -> Vector:
        return Vector(self.__values, self.field)



import random
from ....la2 import PolynomialFraction as pf

id = pf([1], [1])
times_two = pf([2], [1])
plus_5 = pf([1, 5], [1, 0])
x = random.uniform(-100, 100)


# def test_from_string():
#     assert Expression.fromString("x^2")(x) == x**2
#     assert Expression.fromString("x^2+5")(x) == x**2+5
#     assert Expression.fromString("(x+1)^2")(x) == (x+1)**2


def test_addition():
    assert (id+5)(x) == x+5 == (5+id)(x)
    assert (id+id)(x) == 2*x == 2*id(x)


def test_negation():
    assert (-id)(x) == -x == -id(x)


def test_trudiv():
    assert (id/id)(x) == 1
    assert (-id/id)(x) == -1 == (id/-id)(x)


def test_rtruediv():
    assert (1/id)(x) == 1/x == 1/((id/1)(x))
    try:
        (1/id)(0)
        assert False
    except ZeroDivisionError:
        pass


def test_multiplication():
    assert 5*id(x) == (5*id)(x)


def test_pow():
    assert x**2 == id(x)**2 == (id*id)(x) == (id**2)(x)


# def test_rpow():
#     assert (2**id)(x) == 2**x


def test_expression_ception():
    assert id(plus_5(id(id(times_two(plus_5(5)))))) == id(5)**2


def test_equality():
    assert id == id
    assert id(id) == id
    assert plus_5(id) == id(plus_5)
    assert times_two(plus_5) == plus_5+plus_5


# def test_derivative():
#     assert Expression.fromString("x^2").derivetive(x) == 2*x


# def test_antiderivative():
#     assert almost_equal(Expression.fromString("x^2").antiderivative(x), x**3/3)



from __future__ import annotations
from typing import Callable, Union
from ..la1 import Complex, Vector, Field, Matrix, VectorSpace


class BilinearForm:
    # @staticmethod
    # def isBilinearForm() -> bool:
    #     pass

    # @staticmethod
    # def fromInnerProduct(prod) -> BilinearForm:
    #     pass

    # @staticmethod
    # def createSquareBilinearForm() -> BilinearForm:
    #     pass

    @staticmethod
    def fromMatrix(mat: Matrix) -> BilinearForm:
        pass

    def __init__(self, func: Callable[[Vector, Vector], Union[int, float, Complex]], vector_space: VectorSpace) -> None:
        self.func = func
        self.vector_space = vector_space

    @property
    def kernel(self):
        return self.toMatrix().kernel

    @property
    def isSymmetrical(self) -> bool:
        pass

    @property
    def isSquare(self) -> Union[bool, None]:
        return None

    def __call__(self, v, u):
        return self.func(v, u)

    def toMatrix(self) -> Matrix:
        basis = self.vector_space.standard_basis()
        mat = [[0 for i in range(len(basis))] for j in range(len(basis))]
        for i, v in enumerate(basis):
            for j, u in enumerate(basis):
                mat[i][j] = self.func(v, u)
        return Matrix(mat, field=self.vector_space.field)

    def arePerpendicular(self, v1: Vector, v2: Vector) -> bool:
        """
        will return true if result of the bilinear form calculation is 0
        """
        return self.func(v1, v2) == 0


class SquareBilinearForm(BilinearForm):
    @property
    def isSquare(self) -> bool:
        return True

    def __call__(self, v):
        return self.func(v, v)

    def toMatrix(self) -> Matrix:
        pass



from ...la1 import Vector, Complex as c


def test_sum():
    assert sum(Vector([0, 0])) == 0
    assert sum(Vector([1, 2, 3, 4])) == 10


def test_equality():
    assert Vector([0, 0]) == [0, 0]


def test_conjugate():
    assert Vector([0, 1, 2, 3]).conjugate == Vector([0, 1, 2, 3])
    assert Vector([c(1, 1), c(0, 2), c(1, 0)]).conjugate == Vector(
        [c(1, -1), c(0, -2), c(1, 0)])



from setuptools import setup, find_packages
import codecs
import os


VERSION = '0.95'
DESCRIPTION = 'Python linear algebra liabrary'

# Setting up
setup(
    name="python-la",
    version=VERSION,
    author="danielnachumdev (Daniel Nachum)",
    author_email="<danielnachumdev@gmail.com>",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    keywords=['python', 'linear algebra', 'vector', 'matrix',
              'field', 'vector field', 'span', 'linear maps', 'bilinear form', 'inner product', 'linear transformation'],
    classifiers=[
        # "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



from ...la1 import VectorSpace, RealField, ComplexField, DefaultComplexField, Vector


def test_containment():
    Vr3 = VectorSpace(RealField(3))
    Vc3 = VectorSpace(ComplexField(3))
    Vc1 = VectorSpace(ComplexField(1))
    for _ in range(100):
        assert Vr3.random() in Vr3
        assert Vc3.random() in Vc3
        assert Vector([DefaultComplexField.random()], ComplexField(1)) in Vc1



from ...la1 import Complex as c
c  # import sys
# sys.path.insert(1,
#                 "C:\\Users\\danie\\OneDrive\\Desktop\\Code\\VCS\\linear-algebra\\LinearAlgebra")

a = c(1, 1)
b = c(1, 0)


def test_add_RC():
    assert (4 + a).real == 5, "addition with int from the left is not defined"
    assert (4.56 + a).real == 5.56, "addition with int from the left is not defined"


def test_add_CR():
    assert (a + 4).real == 5, "addition with float from the right is not defined"
    assert (c(1.32, 1) +
            4).real == 5.32, "addition with float from the right is not defined"


def test_add_CC():
    assert a+2*a == 3*a
    assert c(1, 0)+c(0, 1) == a
    assert c(-1.5, 0)+c(0, 1.5) == c(-1.5, 1.5)


def test_mul_RC():
    assert 5*a == c(5, 5)


def test_mul_CR():
    assert a*5 == c(5, 5)


def test_mul_CC():
    assert a*a == c(0, 2), "(1+i)(1+i) != 2i"
    assert c(3, 5)*c(5, 3) == c(0, 34), "(3+5i)(5+3i) != 34i"


def test_neg():
    assert -a == c(-1, -1)
    assert -c(-4, -3) == c(4, 3)
    assert -(-c(1, 0)) == c(1, 0)


def test_conjugate():
    assert a.conjugate == a-c(0, 2)
    assert a.conjugate.conjugate == a


def test_inequality():
    assert a is not b
    assert a != b


def test_norm():
    assert (2*b).norm == 4


def test_division():
    assert 1/a == c(1/2, -1/2)
    assert a/2 == c(0.5, 0.5)


def test_pow():
    assert c(1, 1)**2 == c(0, 2)
